/* Copyright 2021 Emanuele Sorce - emanuele.sorce@hotmail.com
 *
 * This file is part of Beru and is distributed under the terms of
 * the GPL. See the file COPYING for full details.
 *
 * This file is part of Sturm Reader and is distributed under the terms of
 * the GPL. See the file COPYING for full details.
 */

import QtQuick 2.12
import QtQuick.Controls 2.5
import QtQuick.Layouts 1.3

Item {
	id: bookStyles

	property bool loading: false
	property bool atdefault: false

	property string textColor
	property string fontFamily
	property alias lineHeight: settingsDialog.lineHeight
	property alias fontScale: settingsDialog.fontScale
	property string background
	property alias margin: settingsDialog.margin
	property real marginv: Math.min(margin, 5)
	property real bumper
	property string pdfBackground
	// real number coefficent that affects pdf quality scale (higher the better)
	property alias pdfQuality: settingsDialog.pdfQuality
	property alias rtl: settingsDialog.rtl

	// color of items such as page number
	property string infoColor: (pdfBackground == "url(.background_paper_invert@30.png)" || pdfBackground == "black") ? "white" : "black"


	BookSettingsDialog {
		id: settingsDialog
	}

	property var defaults: ({
		textColor: "#222",
		fontFamily: "Default",
		lineHeight: 1,
		fontScale: 2,
		background: "url(.background_paper@30.png)",
		pdfBackground: "url(.background_paper@30.png)",
		margin: 0,
		marginv: 0,
		pdfQuality: 1,
		rtl: false,
	})

	//onTextColorChanged: update()  // This is always updated with background
	onFontFamilyChanged: update()
	onLineHeightChanged: update()
	onFontScaleChanged: update()
	onBackgroundChanged: update()
	onPdfBackgroundChanged: update()
	onMarginChanged: update()
	onPdfQualityChanged: update()
	onRtlChanged: update()

	function openDialog() {
		settingsDialog.open();
	}

	function load(styles) {
		loading = true
		textColor = styles.textColor || defaults.textColor
		fontFamily = styles.fontFamily || defaults.fontFamily
		lineHeight = styles.lineHeight || defaults.lineHeight
		fontScale = styles.fontScale || defaults.fontScale
		background = styles.background || defaults.background
		pdfBackground = styles.pdfBackground || defaults.pdfBackground
		margin = styles.margin || defaults.margin
		marginv = styles.marginv || defaults.marginv
		bumper = pictureBook ? 0 : 1
		pdfQuality = styles.pdfQuality || defaults.pdfQuality
		rtl = styles.rtl || defaults.rtl
		loading = false
	}

	function loadForBook() {
		var saved = getBookSetting("styles") || {}
		console.log(JSON.stringify(saved));
		load(saved)
	}

	function asObject() {
		return {
			textColor: textColor,
			fontFamily: fontFamily,
			lineHeight: lineHeight,
			fontScale: fontScale,
			background: background,
			pdfBackground: pdfBackground,
			margin: margin,
			marginv: marginv,
			bumper: bumper,
			pdfQuality: pdfQuality,
			rtl: rtl
		}
	}

	function update() {
		if (loading)
			return

		// book is not loaded
		if(bookWebView.url == "")
			return

		bookLoadingStart()

		// this one below should be improved
		bookWebView.runJavaScript("if(typeof styleManager === 'function') styleManager.updateStyles(" + JSON.stringify(asObject()) + ");");
		if(!setBookSetting("styles", asObject()))
			console.log("Warning! Saving of book styles failed");
		atdefault = (JSON.stringify(asObject()) == JSON.stringify(defaults));
	}

	function resetToDefaults() {
		load({})
		update()
	}

	function saveAsDefault() {
		setSetting("defaultBookStyle", JSON.stringify(asObject()))
		defaults = asObject()
		atdefault = true
	}

	Component.onCompleted: {
		var targetwidth = 60
		var widthgu = width/scaling.dp(8)
		if (widthgu > targetwidth)
			// Set the margins to give us the target width, but no more than 30%.
			defaults.margin = Math.round(Math.min(50 * (1 - targetwidth/widthgu), 30))

		// load defaults
		var saveddefault = getSetting("defaultBookStyle")
		var savedvals = {}
		if (saveddefault)
			savedvals = JSON.parse(saveddefault)
		for (var prop in savedvals)
			if (prop in defaults)
				defaults[prop] = savedvals[prop]

		// Not sure what this does - probabl we don't need it
		//if (savedvals.marginv == undefined && widthgu > targetwidth)
			// Set the vertical margins to be the same as the horizontal, but no more than 5%.
			//defaults.marginv = Math.min(defaults.margin, 5)
	}
}
