/* Copyright 2013-2014 Robert Schroll
 *
 * Copyright 2021 Emanuele Sorce
 *
 * This file is part of Beru and is distributed under the terms of
 * the GPL. See the file COPYING for full details.
 */

#include "epubreader.h"
#include <QJsonObject>
#include <QJsonArray>
#include <QJsonDocument>
#include <QtGui/QImage>
#include <QBuffer>
#include <QDir>
#include <QDebug>
#include <QCryptographicHash>
#include <quazipfile.h>
#include "../qhttpserver/qhttpresponse.h"
#include "../mimetype.h"

QString resolveRelativePath(QString relto, QString path)
{
    int reldirlen = relto.lastIndexOf('/');
    QString reldir = (reldirlen > 0) ? relto.left(reldirlen + 1) : "";
    return QDir::cleanPath(reldir + path);
}

EpubReader::EpubReader(QObject *parent) : QObject(parent)
{
    zip = NULL;
}

bool EpubReader::load(const QString &filename)
{
    if (zip != NULL)
    {
        delete zip;
        zip = NULL;
    }
    _hash = "";
    navhref = "";
    ncxhref = "";
    coverhtml = "";
    spine.clear();
    metadata.clear();
    sortmetadata.clear();

    zip = new QuaZip(filename);
    if (!zip->open(QuaZip::mdUnzip))
    {
        qDebug() << "EpubReader: zip->open() failed";
        delete zip;
        zip = NULL;
        return false;
    }
    if (!parseOPF())
    {
        delete zip;
        zip = NULL;
        return false;
    }
    return true;
}

QString EpubReader::hash()
{
    if (this->_hash != "")
        return this->_hash;

    if (!this->zip || !this->zip->isOpen())
        return this->_hash;

    QByteArray CRCarray;
    QDataStream CRCstream(&CRCarray, QIODevice::WriteOnly);
    QList<QuaZipFileInfo> fileList = this->zip->getFileInfoList();
    foreach (const QuaZipFileInfo info, fileList)
    {
        CRCstream << info.crc;
    }
    this->_hash = QCryptographicHash::hash(CRCarray, QCryptographicHash::Md5).toHex();
    return this->_hash;
}

QString EpubReader::title()
{
    return this->metadata.contains("title") ? this->metadata["title"].toString() : "";
}

QDomDocument *EpubReader::getFileAsDom(const QString &filename)
{
    if (!zip || !zip->isOpen())
        return NULL;

    zip->setCurrentFile(filename);
    QuaZipFile zfile(this->zip);
    if (!zfile.open(QIODevice::ReadOnly))
        return NULL;

    QDomDocument *doc = new QDomDocument();
    if (!doc->setContent(&zfile))
    {
        delete doc;
        zfile.close();
        return NULL;
    }
    zfile.close();
    return doc;
}

void EpubReader::serveComponent(const QString &filename, QHttpResponse *response)
{
    if (!this->zip || !this->zip->isOpen())
    {
        response->writeHead(500);
        response->end("Epub file not open for reading");
        return;
    }

    this->zip->setCurrentFile(filename);
    QuaZipFile zfile(this->zip);
    if (!zfile.open(QIODevice::ReadOnly))
    {
        response->writeHead(404);
        response->end(QString("Could not find \"" + filename + "\" in epub file").toUtf8());
        return;
    }

    response->setHeader("Content-Type", guessMimeType(filename));
    response->writeHead(200);
    // Important -- use write instead of end, so binary data doesn't get messed up!
    response->write(zfile.readAll());
    response->end();
    zfile.close();
}

bool EpubReader::parseOPF()
{
    // Get the container.xml file.
    QDomDocument *container = getFileAsDom("META-INF/container.xml");
    if (container == NULL)
    {
        qDebug() << "Epubreader: couldn't find META-INF/container.xml";
        return false;
    }

    // Find out where the OPF file lives.
    QString contentsfn;
    QDomNodeList nodes = container->elementsByTagName("rootfile");
    // Some ebooks have "odfc:" prefix. I don't know why. It's not in the EPUB standard
    // https://github.com/TronFortyTwo/sturmreader/issues/41
    if (nodes.length() == 0)
        nodes = container->elementsByTagName("odfc:rootfile");

    for (int i = 0; i < nodes.length(); i++)
    {
        QDomElement element = nodes.item(i).toElement();
        if (element.attribute("media-type") == "application/oebps-package+xml")
        {
            contentsfn = element.attribute("full-path");
            break;
        }
    }
    if (contentsfn == "")
    {
        qDebug() << "Epubreader: META-INF/container.xml doesn't contain a valid OPF file";
        return false;
    }

    // Open the OPF file.
    QDomDocument *contents = getFileAsDom(contentsfn);
    if (contents == NULL)
    {
        qDebug() << "Epubreader: Can't find OPF file: " << contentsfn;
        return false;
    }

    // Read the manifest.
    nodes = contents->elementsByTagName("manifest");
    if (nodes.isEmpty())
    {
        qDebug() << "Epubreader: there is no manifest";
        return false;
    }
    QDomElement manifest = nodes.item(0).toElement();
    QHash<QString, QString> idmap;
    nodes = manifest.elementsByTagName("item");
    for (int i = 0; i < nodes.length(); i++)
    {
        QDomElement item = nodes.item(i).toElement();
        idmap[item.attribute("id")] = resolveRelativePath(contentsfn, item.attribute("href"));
        if (item.attribute("properties").split(" ").contains("nav"))
            this->navhref = idmap[item.attribute("id")];
    }

    // Read the spine.
    nodes = contents->elementsByTagName("spine");
    if (nodes.isEmpty())
    {
        qDebug() << "Epubreader: spine empty";
        return false;
    }
    QDomElement spine = nodes.item(0).toElement();
    nodes = spine.elementsByTagName("itemref");
    for (int i = 0; i < nodes.length(); i++)
    {
        QDomElement item = nodes.item(i).toElement();
        this->spine.append(idmap[item.attribute("idref")]);
    }

    // Read the metadata.
    nodes = contents->elementsByTagName("metadata");
    if (nodes.isEmpty())
    {
        // some ebooks use "opf:" prefix for metadata
        // the reason, I ignore it
        nodes = contents->elementsByTagName("opf:metadata");

        if (nodes.isEmpty())
        {
            qDebug() << "Epubreader: metadata empty";
            return false;
        }
    }
    QDomElement dom_metadata = nodes.item(0).toElement();
    nodes = dom_metadata.childNodes();
    for (int i = 0; i < nodes.length(); i++)
    {
        QDomElement item = nodes.item(i).toElement();
        if (!item.isNull() && !item.firstChild().isNull())
        {
            QString name = item.nodeName().split(":").last();
            // subject: append them
            if (name == "subject" && metadata.contains("subject"))
            {
                metadata[name] = metadata[name].toString() + ", " + item.firstChild().nodeValue();
            }
            else
            {
                metadata[name] = item.firstChild().nodeValue();
            }

            // This should work, but doesn't:
            // QString fileas = item.attributeNS("http://www.idpf.org/2007/opf", "file-as");
            QString fileas = item.attribute("opf:file-as");
            if (!fileas.isEmpty())
                sortmetadata[name] = fileas;
        }
    }

    // If this is an Epub3, we've already found the table of contents.  If not,
    // we'll get the Epub2 table of contents.
    if (this->navhref == "")
        this->ncxhref = idmap[spine.attribute("toc")];

    // Look for the HTML file that contains the cover image
    nodes = contents->elementsByTagName("guide");
    if (!nodes.isEmpty())
    {
        QDomElement guide = nodes.item(0).toElement();
        nodes = guide.childNodes();
        for (int i = 0; i < nodes.length(); i++)
        {
            QDomElement reference = nodes.item(i).toElement();
            if (!reference.isNull() && reference.attribute("type") == "cover")
            {
                this->coverhtml = resolveRelativePath(contentsfn, reference.attribute("href"));
                break;
            }
        }
    }
    // If it's not in the guide, guess the first element of the spine
    if (this->coverhtml == "")
        this->coverhtml = this->spine.first();

    return true;
}

QVariantList EpubReader::getContents()
{
    QVariantList res = (this->navhref != "") ? this->parseNav() : this->parseNCX();
    emit contentsReady(res);
    return res;
}

QVariantList EpubReader::parseNav()
{
    QDomDocument *navdoc = this->getFileAsDom(this->navhref);
    QDomNodeList nodes = navdoc->elementsByTagName("nav");
    for (int i = 0; i < nodes.length(); i++)
    {
        QDomElement nav = nodes.item(i).toElement();
        if (nav.attribute("epub:type") == "toc")
        {
            QDomNodeList ols = nav.elementsByTagName("ol");
            if (!ols.isEmpty())
                return this->parseNavList(ols.item(0).toElement());
        }
    }
    return QVariantList();
}

QVariantList EpubReader::parseNavList(QDomElement element)
{
    QVariantList children;
    QDomNodeList nodes = element.childNodes();
    for (int i = 0; i < nodes.length(); i++)
    {
        QDomElement item = nodes.item(i).toElement();
        if (!item.isNull() && item.nodeName() == "li")
        {
            QDomNodeList links = item.elementsByTagName("a");
            if (links.isEmpty())
                continue;
            QDomElement link = links.item(0).toElement();
            QVariantMap entry;
            entry["title"] = link.firstChild().nodeValue();
            entry["src"] = resolveRelativePath(this->navhref, link.attribute("href"));
            QDomNodeList olist = item.elementsByTagName("ol");
            if (!olist.isEmpty())
                entry["children"] = this->parseNavList(olist.item(0).toElement());
            children.append(entry);
        }
    }
    return children;
}

QVariantList EpubReader::parseNCX()
{
    QDomDocument *ncxdoc = this->getFileAsDom(this->ncxhref);
    QDomNodeList nodes = ncxdoc->elementsByTagName("navMap");
    if (nodes.isEmpty())
        return QVariantList();
    return this->parseNCXChildren(nodes.item(0).toElement());
}

QVariantList EpubReader::parseNCXChildren(QDomElement element)
{
    QVariantList children;
    QDomNodeList nodes = element.childNodes();
    for (int i = 0; i < nodes.length(); i++)
    {
        QDomElement node = nodes.item(i).toElement();
        if (!node.isNull() && node.nodeName() == "navPoint")
        {
            QVariantMap entry;
            QDomNodeList labels = node.elementsByTagName("text");
            if (!labels.isEmpty())
                entry["title"] = labels.item(0).firstChild().nodeValue();
            QDomNodeList contents = node.elementsByTagName("content");
            if (!contents.isEmpty())
                entry["src"] = resolveRelativePath(this->ncxhref,
                                                   contents.item(0).toElement().attribute("src"));
            QVariantList child_nav = this->parseNCXChildren(node);
            if (!child_nav.isEmpty())
                entry["children"] = child_nav;
            children.append(entry);
        }
    }
    return children;
}

void EpubReader::serveBookData(QHttpResponse *response)
{
    if (!zip || !zip->isOpen())
    {
        response->writeHead(500);
        response->end("Epub file not open for reading");
        return;
    }

    response->setHeader("Content-Type", guessMimeType("js"));
    response->writeHead(200);
    QJsonDocument json_spine(QJsonArray::fromStringList(spine));
    QJsonDocument json_contents(QJsonArray::fromVariantList(getContents()));
    QJsonDocument json_metadata(QJsonObject::fromVariantMap(metadata));
    QString res = "var bookData = {"
                  "getComponents: function () { return %1; }, "
                  "getContents:   function () { return %2; }, "
                  "getComponent:  function (component) { return { url: component }; }, "
                  "getMetaData:   function (key) { return %3[key]; } }";
    response->write(QString(res.arg(QString(json_spine.toJson()), QString(json_contents.toJson()),
                                    QString(json_metadata.toJson())))
                        .toUtf8());
    response->end();
}

QVariantMap EpubReader::getCoverInfo(int thumbsize, int fullsize)
{
    QVariantMap res;
    if (!zip || !zip->isOpen())
        return res;

    res["title"] = metadata.contains("title") ? metadata["title"] : "ZZZnone";
    res["author"] = metadata.contains("creator") ? metadata["creator"] : "";
    res["authorsort"] = sortmetadata.contains("creator") ? sortmetadata["creator"] : "zzznone";
    res["cover"] = "ZZZnone";

    res["description"] = metadata.contains("description") ? metadata["description"] : "";
    res["subject"] = metadata.contains("subject") ? metadata["subject"] : "";
    res["language"] = metadata.contains("language") ? metadata["language"] : "";

    QDomDocument *coverdoc = getFileAsDom(coverhtml);
    if (coverdoc == NULL)
        return res;

    QString coversrc;
    QDomNodeList images = coverdoc->elementsByTagName("img");
    if (!images.isEmpty())
    {
        coversrc = images.item(0).toElement().attribute("src");
    }
    else
    {
        // Image inside a SVG element
        images = coverdoc->elementsByTagName("image");
        if (!images.isEmpty())
            coversrc = images.item(0).toElement().attribute("xlink:href");
    }
    if (coversrc.isEmpty())
        return res;

    zip->setCurrentFile(resolveRelativePath(this->coverhtml, coversrc));
    QuaZipFile zfile(zip);
    if (!zfile.open(QIODevice::ReadOnly))
        return res;

    QImage coverimg;
    if (!coverimg.loadFromData(zfile.readAll()))
    {
        zfile.close();
        return res;
    }
    zfile.close();
    QByteArray byteArray;
    QBuffer buffer(&byteArray);
    coverimg.scaledToWidth(thumbsize, Qt::SmoothTransformation).save(&buffer, "PNG");
    res["cover"] = "data:image/png;base64," + QString(byteArray.toBase64());
    QByteArray byteArrayf;
    QBuffer bufferf(&byteArrayf);
    coverimg.scaledToWidth(fullsize, Qt::SmoothTransformation).save(&bufferf, "PNG");
    res["fullcover"] = "data:image/png;base64," + QString(byteArrayf.toBase64());
    return res;
}
